# Copyright 1999-2012. Parallels IP Holdings GmbH. All Rights Reserved.
""" Inter-process locking and synchronization primitives (Linux-specific). """
import signal
import errno
import sys
import os
import string
import fcntl

from contextlib import contextmanager

import plesk_log

log = plesk_log.getLogger(__name__)


class TimeoutReachedException(Exception):
    """ Exception that is raised when timeout of some kind was reached
        and some process (code block execution) is terminated as a result.
    """
    def __init__(self, timeout=None):
        self.timeout = timeout

    def __str__(self):
        if isinstance(self.timeout, (int, long, float)) and self.timeout >= 0:
            return 'Timeout of %s seconds has been reached' % self.timeout
        else:
            return 'Timeout has been reached'


class NestingNotSupportedException(Exception):
    """ Exception that is raised when an invalid attempt was made to make
        nested calls of something that doesn't support nesting.
    """
    pass


class AlreadyLockedException(Exception):
    """ Exception that is raised when an attempt to acquire lock failed as it is
        already owned by someone else.
    """
    pass


class InterProcessLockFcntl:
    """ Inter-process synchronization object in form of file-based lock.
        
        Usually you would use a lock via 'with' statement.
        >>> Lock = InterProcessLockFcntl
        >>> with Lock('test'):
        ...     print "Locked"
        ...     Lock('test').get_owner() is not None
        Locked
        True

        Manual acquire/release is also possible.
        >>> l = Lock('test')
        >>> l.acquire()
        >>> os.path.isdir('/var/lock/plesk'), os.path.exists('/var/lock/plesk/test.lock')
        (True, True)
        >>> l.release()
        >>> os.path.exists('/var/lock/plesk/test.lock')
        False

        Try-lock is done via specifying zero timeout.
        >>> with Lock('test'): # doctest: +ELLIPSIS
        ...     with Lock('test', timeout=0):
        ...         print "Unreachable"
        Traceback (most recent call last):
            ...
        AlreadyLockedException: Lock 'test' is already owned by another process with pid ...

        You can also check whether lock is already held by anybody (but this is unreliable).
        >>> print Lock('test').get_owner()
        None

        Nested failed lock attempt don't cause lock to become "abandoned".
        >>> with Lock('test'):
        ...     try:
        ...         with Lock('test', timeout=0):
        ...             print "Unreachable 1"
        ...     except: pass
        ...     try:
        ...         with Lock('test', timeout=0):
        ...             print "Unreachable 2"
        ...     except: pass
        ...     print "Done"
        Done

        You can wait on a lock for a given time.
        >>> with Lock('test'): # doctest: +ELLIPSIS
        ...     with Lock('test', timeout=1):
        ...         print "Unreachable"
        Traceback (most recent call last):
            ...
        AlreadyLockedException: Timeout of 1 seconds has been reached. Lock 'test' is already owned by another process with pid ...
    """
    _allowed_name_chars = frozenset('-_.' + string.ascii_letters + string.digits)
    _locks_dir = '/var/lock/plesk'

    def __init__(self, name=None, timeout=None):
        if not name:
            name = os.path.basename(sys.argv[0])
        if set(name) - self._allowed_name_chars:
            raise ValueError("Invalid name for a lockfile: %s" % name)
        self.name = name
        self.fname = os.path.join(self._locks_dir, self.name + '.lock')
        self.pidfname = os.path.join(self._locks_dir, self.name + '.pid')
        self.lock = None
        self.timeout = timeout

    def _make_locks_dir(self):
        try:
            os.makedirs(self._locks_dir, 0755)
            log.debug("Created a directory for locks: %s", self._locks_dir)
        except OSError, ex:
            if ex.errno != errno.EEXIST:
                raise

    def _lock(self, blocking=True):
        self._make_locks_dir()
        self.lock = open(self.fname, 'w')
        try:
            # Acquire file lock
            try:
                flags = fcntl.LOCK_EX
                if not blocking:
                    flags |= fcntl.LOCK_NB
                fcntl.flock(self.lock, flags)
            except IOError, ex:
                if ex.errno in (errno.EACCES, errno.EAGAIN):
                    # Lock is busy (this check is for not blocking)
                    self.lock.close()
                    self.lock = None
                    return False
                raise

            # Write owner info (lock owner pid)
            with open(self.pidfname, 'w') as pidfile:
                pidfile.write("%s" % os.getpid())

            # Set close-on-exec flag on lock file
            fdflags = fcntl.fcntl(self.lock, fcntl.F_GETFD)
            fcntl.fcntl(self.lock, fcntl.F_SETFD, fdflags | fcntl.FD_CLOEXEC)

            return True
        except:
            self._unlock()
            raise

    def _unlock(self):
        os.unlink(self.pidfname)
        os.unlink(self.fname)
        self.lock.close()
        self.lock = None

    def get_owner(self):
        """ Returns current lock owner id or None if unlocked. This is usually a pid. """
        try:
            with open(self.pidfname, 'r') as pidfile:
                return pidfile.readline().strip()
        except Exception, ex:
            log.debug("Lock '%s' owner is unknown: %s", self.name, ex)
            return None

    def acquire(self, with_timeout=None):
        """ Attempt to acquire lock with optional timeout or in non-blocking (try lock) mode
            if with_timeout == 0. Raises AlreadyLockedException if lock cannot be acquired.
        """
        if with_timeout is None:
            with_timeout = self.timeout

        additional_msg, locked = '', False
        if with_timeout is None:
            locked = self._lock()
        elif with_timeout == 0:
            locked = self._lock(blocking=False)
        else:
            try:
                with timeout(with_timeout):
                    locked = self._lock()
            except TimeoutReachedException, ex:
                additional_msg = "%s. " % ex

        if not locked:
            raise AlreadyLockedException("%sLock '%s' is already owned by another process with pid %s" % 
                                         (additional_msg, self.name, self.get_owner() or "<unknown>"))

    def release(self):
        """ Release previously acquired lock. May raise if lock was not held. """
        self._unlock()

    def __enter__(self):
        self.acquire(self.timeout)

    def __exit__(self, *exc_info):
        try:
            self.release()
        except:
            pass

# Convenience names
FcntlLock = InterProcessLockFcntl
InterProcessLock = InterProcessLockFcntl
Lock = InterProcessLockFcntl


def noop_signal_handler(signum, stack_frame):
    """ No-op signal handler routine to use in signal_handler(). """
    pass

@contextmanager
def signal_handler(signum, handler):
    """ Change signal handler to a new one. Restore original upon 
        exit from a context manager.

        >>> def alarm_handler(signum, stack_frame):
        ...     print "Alarm arrived"
        >>> with signal_handler(signal.SIGALRM, alarm_handler):
        ...     prev = signal.alarm(1)
        ...     print "Alarm armed"
        ...     signal.pause()
        Alarm armed
        Alarm arrived
    """
    original_handler = signal.signal(signum, handler)
    try:
        yield
    finally:
        signal.signal(signum, original_handler)


@contextmanager
def timeout(seconds):
    """ Set a timeout in seconds (integer) for a block of code.
        Raise a TimeoutReachedException if timeout is exceeded.

        This is primarily useful for a code which blocks on I/O, but
        can also be used for any other code. Stack trace would be 
        properly generated.

        Nesting multiple calls to timeout() is not supported and will
        raise a NestingNotSupportedException.

        >>> with timeout(5):
        ...     print "Timeout was not reached"
        Timeout was not reached

        >>> try:
        ...     with timeout(1):
        ...         signal.pause()
        ...         print "Will not be printed"
        ... except TimeoutReachedException, ex:
        ...     print "OK: %s" % ex
        OK: Timeout of 1 seconds has been reached

        >>> from fcntl import flock, LOCK_EX
        >>> fname = 'test_lock.lock'
        >>> try:
        ...     with timeout(1):
        ...         f1, f2 = open(fname, 'w'), open(fname, 'w')
        ...         flock(f1, LOCK_EX)
        ...         print "First lock grabbed"
        ...         flock(f2, LOCK_EX)
        ...         print "Second lock grabbed"
        ... except TimeoutReachedException, ex:
        ...     print "OK: %s" % ex
        ... finally:
        ...     f1.close(); f2.close()
        ...     os.unlink(fname)
        First lock grabbed
        OK: Timeout of 1 seconds has been reached

        >>> with timeout(3):
        ...     with open(fname, 'r') as f:
        ...         pass
        Traceback (most recent call last):
            ...
        IOError: [Errno 2] No such file or directory: 'test_lock.lock'

        >>> def crunch():
        ...     s = 0
        ...     while True:
        ...         s += 1
        >>> with timeout(1):
        ...     crunch() # Will timeout inside this function
        Traceback (most recent call last):
            crunch() # Will timeout inside this function
            ### I.e. stack trace will be correct. Note that this is not really checked by doctest :(
        TimeoutReachedException: Timeout of 1 seconds has been reached

        >>> with timeout(2):
        ...     with timeout(1):
        ...         pass
        Traceback (most recent call last):
            ...
        NestingNotSupportedException: timeout() cannot be nested. Previously set timeout was removed.
    """
    def timeout_handler(signum, stack_frame):
        if signum == signal.SIGALRM:
            raise TimeoutReachedException(seconds)

    if seconds <= 0:
        raise ValueError("Invalid timeout value: %d" % seconds)

    with signal_handler(signal.SIGALRM, timeout_handler):
        try:
            if signal.alarm(seconds) != 0:
                raise NestingNotSupportedException("timeout() cannot be nested. "
                                                   "Previously set timeout was removed.")
            yield
        except EnvironmentError, ex:
            if ex.errno == errno.EINTR:
                raise TimeoutReachedException(seconds)
            raise
        finally:
            signal.alarm(0)


if __name__ == '__main__':
    import doctest
    print "Running doctest (this may take some time since includes timeout() testing)"
    sys.exit( doctest.testmod().failed )

# vim: ts=4 sts=4 sw=4 et :
